#ifndef _NTSC_VIDEO_H_
#define _NTSC_VIDEO_H_

#include <plib.h>
#include <xc.h> // need for pps
#include <stdio.h>

#include "ascii_characters.h"

// The DMA channels
#define DMAchn1 1
#define DMApri0 0

// video timing
#define line_cycles 1905 // 63.5 uSec at 30 MHz Fpb, prescaler=1
#define us_5_cycles 150 // 5 uSec at 30 MHz Fpb, prescaler=1
#define us_13_cycles 390 // beginning of the video DMA burst -- use for centering
#define us_11_cycles 330 // beginning of the video DMA burst -- use for centering
#define us_10_cycles 300 // beginning of the video DMA burst -- use for centering

// video active lines -- 200 total
#define image_start 20
#define image_end image_start+200

#define top 0
#define left 7
#define right 248
#define bottom 199

#define DRAW_BLACK 0
#define DRAW_WHITE 1
#define DRAW_NEGATE 2

//=== fixed conversion macros =========================================
// for 16.16 fixed point
typedef signed int fix ;
#define int2fix(a)   (((fix)(a))<<16)            //Convert char to fix. a is a char
#define fix2int(a)   ((signed int)((a)>>16))    //Convert fix to int. a is an int
#define float2fix(a) ((fix)((a)*65536.0))         //Convert float to fix. a is a float
#define fix2float(a) (((float)(a))/65536.0)       //Convert fix to float. a is an int
#define multfix(a,b) ((fix)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16.16

#define fixpercent float2fix(0.01)

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

// helper functions to draw
void video_pt(int x, int y, char c);
void video_line(int x1, int y1, int x2, int y2, char c);
char video_state(int x, int y);
void video_putchar(int x, int y, int c);
void video_string(int x, int y, char *str);
void video_big_putchar(int x, int y, int c);
void video_bold_string(int x, int y, char *str);

void init_ntsc_video();
void test_video_buffer();

// more helper functions to draw

#define OPT_OVERWRITE 0
#define OPT_OVERLAY 1
#define OPT_NEGATE 2
#define OPT_REMOVE 3
// copy a block of image data to a specific point on screen
// opt is for options: 
// OPT_OVERWRITE: overwrites the data below
// OPT_OVERLAY: overlays the data
// OPT_NEGATE: negates the data
// OPT_REMOVE: removes data
void video_block(int x, int y, int width, int height, int* arr, char opt);
void video_reverseblock(int x, int y, int width, int height, int* arr, char opt);

// removes a block of char on given position
void video_removechar(int x, int y, char c);
void video_removestring(int x, int y, char *str);

void video_delchar(int x, int y, char c);
void video_delstring(int x, int y, char *str);

// below are set of gfx functions ported from tft_gfx library
void video_drawRect(short x, short y, short w, short h, char opt);
void video_fillRect(short x0, short y0, short w, short h, char opt);
void video_drawCircle(short x0, short y0, short r, char opt);
void video_drawCircleHelper(short x0, short y0, short r, unsigned char cornername, char opt);
void video_fillCircle(short x0, short y0, short r, char opt);
void video_fillCircleHelper(short x0, short y0, short r, unsigned char cornername, short delta, char opt);
void video_drawTriangle(short x0, short y0, short x1, short y1, short x2, short y2, char opt);
void video_fillTriangle(short x0, short y0, short x1, short y1, short x2, short y2, char opt);
void video_drawRoundRect(short x0, short y0, short w, short h, short r, char opt);
void video_fillRoundRect(short x0, short y0, short w, short h, short r, char opt);
void video_drawBitmap(short x, short y, const unsigned char *bitmap, short w, short h, char opt);


// bezier curves and more complicated drawings
// x0,y0 is the starting point
// x1,y1 is the pivot point
// x2,y2 is the end point
void video_drawBezier(int x0, int y0, int x1, int y1, int x2, int y2, char opt);
int video_drawBezierHelper(int n1, int n2, fix perc);

void video_clr();

#endif