#include "game_manager.h"

// sine table
#define SINE_TABLE_SIZE 256
fix sine_table[SINE_TABLE_SIZE];

// variables storing valuable information
extern volatile int time_tick_60_hz;

char rope_side = 0;
int rotation = 0;

int bez_y;
float v_timer = 1.00;
int difficulty = 1; // scales from 1 to 100
int i_rope = 128, i_old = 128;
int v = 12;
char game_over_status = 0, check_game_over_flag = 0;
char life_str[10];
int old_bezier_pos, bezier_pos, bezier_target_pos;
char reset_status = 1;
char life = 3;

// ========================= tooltip sprite =============================
// the purpose of the tooltip sprite is to separate all the helpful text
// that will float during game with the game itself.
// it keeps track of all the floating messages on the fly
// also animates them to death
// it contains tool tip such as speed up, random speed, move, etc.

char tooltip_state = 0;
char tooltip_enabled = 0;
char tooltip_conf_enabled = 0;
int tooltip_time = 0;
int tooltip_prev_time = 0;
char tooltip_str[15];
fix tooltip_pos_y;

void tooltip_update(sprite_t* s, int dt){
    static short tooltip_pos_x;
    // quick conf
    if (tooltip_conf_enabled){
        config_hit(2);
        DmaChnEnable(DMAhit);
        switch (tooltip_state) {
            case TOOLTIP_SPEED:
                tooltip_pos_x = 105;
                tooltip_pos_y = int2fix((bottom+top)/4 + 20);
                sprintf(tooltip_str, "SPEED UP!");
                video_string(tooltip_pos_x,fix2int(tooltip_pos_y),tooltip_str);
                break;
            case TOOLTIP_RANDOM:
                tooltip_pos_x = 90;
                tooltip_pos_y = int2fix((bottom+top)/4 + 20);
                sprintf(tooltip_str, "RANDOM SPEED!");
                video_string(tooltip_pos_x,fix2int(tooltip_pos_y),tooltip_str);
                break;
            case TOOLTIP_MOVE:
                tooltip_pos_x = 115;
                tooltip_pos_y = int2fix((bottom+top)/4 + 20);
                sprintf(tooltip_str, "MOVE!");
                video_string(tooltip_pos_x,fix2int(tooltip_pos_y),tooltip_str);
                break;
            case TOOLTIP_DIE:
                tooltip_pos_x = 105;
                tooltip_pos_y = int2fix((bottom+top)/4 + 20);
                sprintf(tooltip_str, "LIFE -1");
                video_string(tooltip_pos_x,fix2int(tooltip_pos_y),tooltip_str);
                break;
            default:
                break;
        }
        tooltip_time = time_tick_60_hz;
        tooltip_enabled = 1;
        tooltip_conf_enabled = 0;
    }
    // if configured, advance the counter
    if (tooltip_prev_time != time_tick_60_hz){
        if (tooltip_enabled){
                video_delstring(tooltip_pos_x, fix2int(tooltip_pos_y), tooltip_str);
                tooltip_pos_y -= float2fix(0.5);
                video_string(tooltip_pos_x,fix2int(tooltip_pos_y),tooltip_str);
            if (time_tick_60_hz - tooltip_time >= 45) {
                tooltip_enabled = 0;
                config_hit(1);
                video_delstring(tooltip_pos_x, fix2int(tooltip_pos_y), tooltip_str);
            }
        }
    }
    tooltip_prev_time = time_tick_60_hz;
}

void tooltip_init(sprite_t* s, int dt){
    s->update_func = tooltip_update;
}

sprite_t tooltip = {0,0,0,0,0,0,tooltip_init,0,0};


// ======================= character sprite ==============================
void character_update(sprite_t* s, int dt){
    // update the two characters on the sides
    static int old_frame = 2;
    video_block(s->pos.x, s->pos.y+MAN_ACTIVE_AREA_TOP, s->width, s->height-MAN_ACTIVE_AREA_TOP-MAN_ACTIVE_AREA_BOTTOM, (int*)((s->data)+s->height*old_frame+MAN_ACTIVE_AREA_TOP), OPT_REMOVE);
    video_reverseblock(s->pos.x-59, s->pos.y+MAN_ACTIVE_AREA_TOP, s->width, s->height-MAN_ACTIVE_AREA_TOP-MAN_ACTIVE_AREA_BOTTOM, (int*)((s->data)+s->height*old_frame+MAN_ACTIVE_AREA_TOP), OPT_REMOVE);
    old_frame = s->current_frame;
    video_block(s->pos.x, s->pos.y+MAN_ACTIVE_AREA_TOP, s->width, s->height-MAN_ACTIVE_AREA_TOP-MAN_ACTIVE_AREA_BOTTOM, (int*)((s->data)+s->height*s->current_frame+MAN_ACTIVE_AREA_TOP), OPT_OVERLAY);
    video_reverseblock(s->pos.x-59, s->pos.y+MAN_ACTIVE_AREA_TOP, s->width, s->height-MAN_ACTIVE_AREA_TOP-MAN_ACTIVE_AREA_BOTTOM, (int*)((s->data)+s->height*s->current_frame+MAN_ACTIVE_AREA_TOP), OPT_OVERLAY);
}


void character_init(sprite_t* s, int dt){
    s->update_func = character_update;
    vector2d pos = {170, 72};
    s->pos = pos;
    s->data = MAN_BUF;
    s->frames = 4;
    s->current_frame = 2; // starting from 0
    s->width = 32;
    s->height = MAN_SIZE/s->frames;
    video_block(s->pos.x, s->pos.y, s->width, s->height, (int*)((s->data)+s->height*s->current_frame), OPT_OVERLAY);
    video_reverseblock(s->pos.x-59, s->pos.y, s->width, s->height, (int*)((s->data)+s->height*s->current_frame), OPT_OVERLAY);
}

sprite_t character = {0,0,0,0,0,0,character_init,0,0};

// =========================== rope sprite ==============================
// helpful functions
void rope_difficulty_config(float* timer_interval, int* v, int* difficulty, int rotation){
    static int rnd_val = 0;
    // automatically advances the level and adjusts difficulty
    if (rotation % 40 == 0 && rotation != 0) {
        // multiple of 40 will get random curve x position
        bezier_target_pos = rand() % 120 + (right+left)/2 - 60;
        *timer_interval = MAX(1.00-(float)(*difficulty)*0.03, 0.50);
        *v=MIN(11+(*difficulty)*2, 45);
        tooltip_state = TOOLTIP_MOVE;
        tooltip_conf_enabled = 1;
        config_speed(v_timer);
    } else if (rotation % 30 == 0 && rotation != 0){
        // multiple of 30 will get random speed
        tooltip_state = TOOLTIP_RANDOM;
        tooltip_conf_enabled = 1;
        rnd_val = rand() % (11+(*difficulty)*2);
        *v=MAX(5, rnd_val);
        *timer_interval = MAX(1.25-(float)(rnd_val)*0.01, 0.50);
        config_speed(v_timer);
    } else if (rotation % 40 >=10 && rotation % 30 < 10 && rotation > 10){
        // random speed
        rnd_val = rand() % (11+(*difficulty)*2);
        *v=MAX(5, rnd_val);
        *timer_interval = MAX(1.25-(float)(rnd_val)*0.01, 0.50);
        config_speed(v_timer);
    } else if (rotation % 10 == 0 && rotation != 0) {
        // multiple of 10 will get speed up
        (*difficulty)++;
        *timer_interval = MAX(1.00-(float)(*difficulty)*0.03, 0.50);
        *v=MIN(11+(*difficulty)*2, 45);
        tooltip_state = TOOLTIP_SPEED;
        tooltip_conf_enabled = 1;
        config_speed(v_timer);
    }
}

char check_end_game(sprite_t* rope0, sprite_t* ball0){
    // check for end game condition
    if (ball0->pos.y >= rope0->pos.y+rope0->height/2-18 || ball0->pos.x <= bezier_pos-50 || ball0->pos.x >= bezier_pos+50){
        return 1;
    } else {
        return 0;
    }
}

void approach_by_one(int* out, int* tgt){
    // helper function to approach the source by one step closer to the destination
    if (*tgt > *out){
        (*out) ++;
    } else if (*tgt < *out){
        (*out) --;
    }
}

void rope_update(sprite_t* s, int dt){
    // update rope
    // erase old one first
    video_drawBezier(s->pos.x - s->width/2, s->pos.y, old_bezier_pos, bez_y, s->pos.x + s->width/2, s->pos.y, DRAW_BLACK);
    
    // approach the current bezier postion with the target postion
    approach_by_one(&bezier_pos, &bezier_target_pos);
    old_bezier_pos = bezier_pos;
    
    // compute the bezier y position
    bez_y = fix2int(multfix(int2fix(s->height),sine_table[i_rope])) + s->pos.y - s->height/5;
    // velocity of the rope, simulating gravity
    if (i_rope <= 96 || i_rope >= 224) {
        i_rope += v;
    } else {
        i_rope += (v*7/10);
    }
    if (i_rope >= SINE_TABLE_SIZE) {
        i_rope = 0;
    }
    // detecting which side the rope is at, also configs difficulty when the side changes
    if (i_rope >= 64 && i_old < 64){
        rope_side ^= 1;
        rotation++;
        // increase difficulty
        rope_difficulty_config(&v_timer, &v, &difficulty, rotation);  
        check_game_over_flag = 1;
        character.current_frame = 3;
    } else {
        check_game_over_flag = 0;
    }
    // flips the side and changes character frame
    if (i_rope >= 192 && i_old < 192){
        rope_side ^= 1;
        character.current_frame = 0;
    }
    // changes character frame
    if (i_rope >= 224 && i_old < 224){
        character.current_frame = 1;
    }
    if (i_rope >= 160 && i_old < 160){
        character.current_frame = 1;
    }
    if (i_rope >= 32 && i_old < 32){
        character.current_frame = 2;
    }
    if (i_rope >= 96 && i_old < 96){
        character.current_frame = 2;
    }
    
    i_old = i_rope;
    // draw the bezier curve again
    video_drawBezier(s->pos.x - s->width/2, s->pos.y, bezier_pos, bez_y, s->pos.x + s->width/2, s->pos.y, DRAW_WHITE);
}

void rope_init(sprite_t* s, int dt){
    int k;
    s->width = 100;
    s->height = 40;
    vector2d pos = {(right+left)/2,(top+bottom)/2};
    s->pos = pos;
    s->current_frame = 0;
    s->data = NULL;
    s->frames = 0;
    s->update_func = rope_update;
    
    // init sine table
    for (k = 0; k < SINE_TABLE_SIZE; k++) {
        sine_table[k] = float2fix(sin((float)k*6.283/(float)SINE_TABLE_SIZE));
    }
    
    if (reset_status) {
        bez_y = s->pos.y-s->height;
        old_bezier_pos = s->pos.x;
        bezier_pos = s->pos.x;
        bezier_target_pos = s->pos.x;
    }
}

sprite_t rope = {0,0,0,0,0,0,rope_init,0,0};

// ============================= ball sprite ============================
int key_press_timer = 0;
char w_str[2];
char dot_str[2];
int eye_data[] = {0x50000000};
void ball_update(sprite_t* s, int dt){
    static fix f_y = 0;
    static fix v_y = 0;
    static char old_key;
    video_drawRoundRect(s->pos.x-s->width/2,s->pos.y-s->height/2,s->width,s->height,3,DRAW_BLACK);
    video_delstring(s->pos.x, s->pos.y, w_str);
    video_block(s->pos.x, s->pos.y-2, 4, 1, eye_data, OPT_REMOVE);
    // poll for the key
    if (getbuttondown(KEY_LEFT)){
        s->pos.x -= 2;
    }
    
    if (getbuttondown(KEY_RIGHT)){
        s->pos.x += 2;
    }
    // sudden thrust when key A pressed
    if (getbuttondown(KEY_A)){
        if (old_key == 0){
            if ((time_tick_60_hz-key_press_timer)>=20){
                DmaChnEnable(DMAhit);
                f_y = float2fix(5.2);
                v_y = 0;
                key_press_timer = time_tick_60_hz;
            }
        } else {
            f_y = float2fix(-1.5);
        }
        old_key = 1;
    } else {
        f_y = float2fix(-1.5);
        old_key = 0;
    }
    // basic physics engine
    v_y = MAX(v_y+f_y, int2fix(-6));
    s->pos.y = MIN(rope.pos.y+rope.height/2 - 14, s->pos.y-fix2int(v_y));
    s->pos.y = MAX(rope.pos.y-rope.height, s->pos.y);
    s->pos.x = MAX(rope.pos.x-rope.width/2+2, s->pos.x);
    s->pos.x = MIN(rope.pos.x+rope.width/2-2, s->pos.x);
    video_fillRoundRect(s->pos.x-s->width/2,s->pos.y-s->height/2,s->width,s->height,3,DRAW_BLACK);
    video_string(s->pos.x, s->pos.y, w_str);
    video_block(s->pos.x, s->pos.y-2, 4, 1, eye_data, OPT_OVERLAY);
    video_drawRoundRect(s->pos.x-s->width/2,s->pos.y-s->height/2,s->width,s->height,3,DRAW_WHITE);
}

void ball_init(sprite_t* s, int dt){
    s->width =16;
    s->height =18;
    if (reset_status){
        vector2d pos = {(right+left)/2,(top+bottom)/2};
        s->pos = pos;
    }
    s->current_frame = 0;
    s->data = NULL;
    s->frames = 0;
    s->update_func = ball_update;
    sprintf(w_str, "w");
}

sprite_t ball = {0,0,0,0,0,0,ball_init,0,0};

// ============================= rope scene =============================
extern unsigned char controller_status;
char ctrl_string[10] ;
char rot_str[15];
void scene_rope_update(scene_t* s, int dt){
    int i;
    
    // adjust the rendering order depending on the layer order.
    // draws the background line to prevent it being erased by the ball
    video_line(rope.pos.x-(rope.width/2)-5, rope.pos.y+rope.height/2 - 17, rope.pos.x+(rope.width/2), rope.pos.y+rope.height/2 - 17, DRAW_WHITE);
    
    if (!rope_side) {
        s->sprites[0]->update_func(s->sprites[0], 0);
        s->sprites[1]->update_func(s->sprites[1], 0);
    } else {
        s->sprites[1]->update_func(s->sprites[1], 0);
        s->sprites[0]->update_func(s->sprites[0], 0);
    }
    
    // update its children sprites first
    for (i = 2; i < s->nsprite; i++){
        s->sprites[i]->update_func(s->sprites[i], 0);
    }
    
    sprintf(ctrl_string, "%03d %03d", bezier_pos, ball.pos.x);
    video_string(left+10, bottom-10, ctrl_string);
    
    sprintf(rot_str, "Score: %03d", rotation);
    video_string(left+14, top+24, rot_str);
    
    sprintf(life_str, "Lives: %d", life);
    video_string(right-59, top+24, life_str);
    
    if (life == 0){
        config_speed(1.00);
        game_manager_load_scene(4);
    }
    
    if (check_game_over_flag){
        game_over_status = check_end_game(&rope, &ball);
    }
    
    if (game_over_status){
        life --;
        tooltip_state = TOOLTIP_DIE;
        tooltip_conf_enabled = 1;
        game_over_status = 0;
    }
    
    if (getbuttondown(KEY_START)){
        config_speed(1.00);
        game_manager_load_scene(3);
    }
}

sprite_t* rope_scene_sprites[4];

void scene_rope_init(scene_t*s, int dt){
    int i;
    
    video_clr(); // clear screen
    
    if (s->reset == 1){
        // reset all the parameters
        difficulty = 1;
        rotation = 0;
        v_timer = 1.00;
        v = 12;
        i_rope = 128;
        i_old = 128;
        s->reset = 0;
        game_over_status = 0;
        tooltip_enabled = 0;
        reset_status = 1;
        life = 3;
    }
    
    s->nsprite = 4;
    
    s->sprites = rope_scene_sprites;
    s->sprites[0] = &rope;
    s->sprites[1] = &ball;
    s->sprites[2] = &tooltip;
    s->sprites[3] = &character;

    s->update_func = scene_rope_update;
    
    // init all the sprites in the scene
    for (i = 0; i < s->nsprite; i++){
        s->sprites[i]->init_func(s->sprites[i], 0);
    }
    
    reset_status = 0;
    
    // init the scene here
    // Draw the screen boundaries
    video_line(left+10,top+20, right-10,top+20, 1); // top
    video_line(right-10,top+20, right-10,bottom-20, 1); // right
    video_line(right-10,bottom-20, left+10,bottom-20, 1); // bottom
    video_line(left+10,top+20, left+10,bottom-20 ,1); // left
    
    video_line(60, rope.pos.y+rope.height/2 - 17, 205, rope.pos.y+rope.height/2 - 17, DRAW_WHITE);
    
    video_line(45, rope.pos.y+rope.height/2+14, 195, rope.pos.y+rope.height/2+14, DRAW_WHITE);
    // audio config
    config_bgm(0);
    config_hit(1);
    
    config_speed(v_timer);
}

scene_t scene_rope = {scene_rope_init, NULL, NULL, 0, 1};